/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "lookup_stage.hh"
#include "write_stage.hh"
#include "chan_list.hh"
#include "driver_table.hh"
#include "frame_source.hh"
#include "source_histo.hh"
#include "sysexcept.hh"
#include <iostream>
#include <sstream>
#include <stdexcept>

using namespace sends;
using namespace std;

//======================================  Construct lookup stage
lookup_stage::lookup_stage(int nThread, int nTrans)
  : seda_stage(nThread, nTrans)
{
}

//======================================  Construct lookup stage
lookup_stage::~lookup_stage(void) {
}

//======================================  Return stage name
const char*
lookup_stage::stage_name(void) const {
    return "lookup";
}

//======================================  Look up requested channels
void
lookup_stage::build_source_list(transaction* trans, source_list& svec) {

    //----------------------------------  Initialize the output vector.
    index_type N = trans->getNChan();
    svec.resize(N);
    for (index_type i=0; i<N; ++i) svec[i].clear();

    //----------------------------------  Initialize the output vector.
    gps_type start = trans->getStartGPS();
    source_histo hist;
    index_type iter_count = 0;
    for (index_type nSourced=0; nSourced < N; ) {

	//------------------------------  Histogram data sources
	for (index_type i=0; i<N; ++i) {
	    chan_index chan = trans->chan(i).index();
	    if (svec[i].empty()) hist.bump(Channel_List[chan], start);
	}
	if (hist.empty()) {
	    throw runtime_error("Unable to find source for data channels");
	}

	//------------------------------  Assign best source
	string src = hist.first()._name;
	hist.clear();
	debug_msg(dlvl_detail, trans->ident(), 
		  string("Assign channels to: ")+src);
	for (index_type i=0; i<N; ++i) {
	    chan_index chan = trans->chan(i).index();
	    if (svec[i].empty() && Channel_List[chan].is_source(src, start)) {
		debug_msg(dlvl_detail, trans->ident(), 
			  string("sourced: ")+Channel_List[chan].name());
		svec[i] = src;
		nSourced++;
	    }
	}
	hist.veto(src);
	iter_count++;
    }
    if (debug(dlvl_detail)) {
	ostringstream msg;
	msg << "number of iterations in lookup: " << iter_count << endl;
	debug_msg(dlvl_detail, trans->ident(), msg.str());
    }
}

//======================================  Look up requested channels
void
lookup_stage::filter(transaction* trans) {
    trans->set_stage(300);

    //----------------------------------  Start the lookup stage processing
    switch (trans->command()) {

    //----------------------------------  Process online data request
    case transaction::kGetOnlineData:
	try {
	    //--------------------------  Check access
	    index_type N = trans->getNChan();
	    for (index_type i=0; i<N; ++i) {
		chan_index inx = trans->chan(i).index();
		const chan_entry& chan(Channel_List[inx]);
		if (chan.type() != chan_type::kOnline) {
		    throw runtime_error(string("Channel not online: ") 
					+ chan.name());
		}

		if (!chan.test_access(trans->ref_user().access())) {
		    trans->chan(i).set_stat(-2);
		    //throw runtime_error(string("Channel protected: ") 
		    //		+ Channel_List[inx].name());
		}
	    }

	    //------------------------------  Make sure there's a source
	    source_list svec(N);
	    build_source_list(trans, svec);

	    //------------------------------  Reply for valid data.
	    trans->put_response(DAQD_OK);

	    //------------------------------  Start the drivers...
	    debug_msg(dlvl_detail, trans->ident(), "Source list follows:");
	    for (index_type i=0; i < N; i++) {
		if (svec[i].empty()) continue;
		string name = svec[i];
		const loc_list* ll=Source_List.list_ptr(loc_list::kOnline,name);
       		driver_api& driver = TheDriverTable[ll->back().driver()];
		if (debug(dlvl_detail)) cout << svec[i];
		for (index_type j=i; j<N; ++j) {
		    if (svec[j] == name) {
			chan_index cinx = trans->chan(j).index();
			driver.request_channel(cinx);
			trans->chan(j).set_driver(&driver);
			if (debug(dlvl_detail)) {
			    cout << "  " << Channel_List[cinx].name();
			}
			svec[j].clear();
		    }
		}
		cout << endl;
	    }

	    //------------------------------  Send a response...
	    debug_msg(dlvl_meta, trans->ident(), "Send an early response");
	    if (trans->get_state() == transaction::ws_initial) {
		debug_msg(dlvl_detail, trans->ident(), "writing trans ID");
		socket_api& s(trans->ref_socket());
		int rc = s.put_hex(trans->ident(), 8);
		if (rc >= 0) {
		    debug_msg(dlvl_detail, trans->ident(), "transfer type=0");
		    rc = s.put_int(0); // "offline" flag
		}
		if (rc < 0) {
		    string msg = "lookup_stage::filter: data write error in ";
		    msg += s.where();
		    throw system_error(msg, s.sys_errno());
		}
		trans->set_state(transaction::ws_response);
	    }

	    //------------------------------  Get the start time...
	    debug_msg(dlvl_meta, trans->ident(), "Get the start time");
	    gps_type start = 0;
	    for (index_type i=0; i < N; i++) {
		driver_api* driver = trans->chan(i).driver();
		gps_type ti = driver->online_gps();
		if (!ti) {
		    trans->chan(i).set_stat(-1);
		    driver->release_channel(trans->chan(i).index());
		    trans->chan(i).set_driver(0);
		} else if (ti > start) {
		    start = ti;
		}
	    }

	    if (debug(dlvl_transaction)) {
		ostringstream out;
		out << "Online start time is: " << start;
		debug_msg(dlvl_transaction, trans->ident(), out.str());
	    }

	    //------------------------------  Reserve the source segments
	    if (trans->getStartGPS() == 0) {  // start=0 if it hasn't been set
		gps_type stop   = trans->getStopGPS();
		gps_type stride = trans->stride();
		if (stop != 0 && stop <= start) stop += start;
		trans->setTime(start, stop, stride);
		trans->start_fill(start);
	    }

	    //------------------------------  Pass transaction to next stage
	    put_output(trans);

	//----------------------------------  Error in lookup processing.
	} catch (exception& e) {
	    trans->put_response(DAQD_NOT_FOUND);
	    trans->set_stage(320);
	    reject(trans, e.what());
	}
	break;

    //--------------------------------------  Check access
    case transaction::kGetData:
	try {
	    index_type N = trans->getNChan();
	    for (index_type i=0; i<N; ++i) {
		chan_index inx = trans->chan(i).index();
		if (!Channel_List[inx].test_access(trans->ref_user().access())){
		    trans->chan(i).set_stat(-2);
		    //throw runtime_error(string("Channel protected: ") 
		    //			+ Channel_List[inx].name());
		}
	    }

	    //------------------------------  Make sure there's a source
	    source_list svec(N);
	    build_source_list(trans, svec);

	    //------------------------------  Reply for valid data.
	    trans->put_response(DAQD_OK);

	    //------------------------------  Reserve cached channels
	    trans->start_fill(trans->getStartGPS());
	    trans->reserve_segments();

	    //------------------------------  Pass transaction to next stage
	    put_output(trans);

	//----------------------------------  Error in lookup processing.
	} catch (exception& e) {
	    trans->put_response(DAQD_NOT_FOUND);
	    trans->set_stage(320);
	    reject(trans, e.what());
	}
	break;

    //--------------------------------------  Non-data commands
    default:
	//----------------------------------  Pass transaction to next stage
	trans->set_stage(330);
	put_output(trans);
	break;
    }
}
